import java.util.*;
/**
 * Represents the checkerboard as a 33 length string, where chars 1-32 
 * correspond to the checkerboard and the 0 position is left empty.
 *  
 * 1) Provides a boolean to indicate which color's turn it is as well as
 * storage for the move made by the agent.  
 * 2)Provides a method to calculate a utility value of the board based 
 * on our heuristic.
 * 3)Provides a method to calculate all possible successor boards for
 * the color who's turn it is.
 * 
 * @author Cheryl Yager
 * @author Umar K. Munroe
 * @version 1.0
 */
public class State
{

	private String board;


	public static final int BOARD_SIZE = 33;

	private static final char RED = 'r';

	private static final char RED_KING = 'R';

	private static final char WHITE = 'w';

	private static final char WHITE_KING = 'W';

	private static final char BLANK = ' ';

	private boolean itIsRedTurn;

	private boolean turnSet = false;

	private static final int START_POSITION = 1;

	private static boolean agentIsRed = true;

	private int numJumps = 0;

	private Integer[] movesMade;

	/**
	 * Create a state with initial board.
	 * 
	 * @param redIsFirst is true if the state should indicate that it is red's
	 *            move.
	 */
	public State(boolean redIsFirst)
	{
		char[] initialPositions = new char[BOARD_SIZE];
		setTurn(redIsFirst);
		for (int i = 1; i <= 12; i++)
		{
			initialPositions[i] = RED;
		}
		for (int i = 13; i <= 20; i++)
		{
			initialPositions[i] = BLANK;
		}
		for (int i = 21; i <= 32; i++)
		{
			initialPositions[i] = WHITE;
		}
		
		/////////////////////////////////
		/*
		for(int i=0; i<BOARD_SIZE; i++)
		{
			initialPositions[i] = BLANK;
		}
		*/
		///////////////////////////////////
		board = new String(initialPositions);

		movesMade = new Integer[0];
	}

	/**
	 * Create a state with the input board. The state does not indicate who's
	 * move it is until it is set by the setTurn(x) method. Primarily used by
	 * the agent.
	 * 
	 * @param theBoard the location of all the pieces
	 * @param theNumJumps the number of jumps that resulted in the input board
	 * @param theMoves the start, intermediate, and end moves that resulted in
	 *            the input board
	 */
	public State(String theBoard, int theNumJumps, Integer[] theMoves)
	{
		// initialise instance variables
		//utilitySet = false;
		board = new String(theBoard);
		numJumps = theNumJumps;
		movesMade = theMoves;
	}

	/**
	 * Create a state with the input board. The state does not indicate who's
	 * move it is until it is set by the setTurn(x) method. Primarily used by
	 * the opponent.
	 * 
	 * @param theBoard the location of all the pieces
	 */
	public State(String theBoard)
	{
		this(theBoard, 0, new Integer[0]);
	}

	/**
	 * @return all legal moves of the color who's turn this state indicates.
	 */
	public State[] getSuccessors()
	{
		boolean redMove = isRedTurn();
		ArrayList validMoves = new ArrayList();
		State[] moves;
		int maxJumps = 0;
		for (int i = START_POSITION; i < BOARD_SIZE; i++)
		{
			//only look for the valid moves of the color who's turn it is
			if ((redMove && isRed(i)) || (!redMove && isWhite(i)))
			{
				moves = getValidMoves(i, false);
				//find the max number of jumps seen for a single checker
				for (int j = 0; j < moves.length; j++)
				{
					if (moves[j].numJumps > maxJumps)
					{
						maxJumps = moves[j].numJumps;
					}
				}
				//go back through all moves for the single checker
				for (int j = 0; j < moves.length; j++)
				{
					//if a jump is available, valid moves are the jumps ONLY
					if (maxJumps > 0 && moves[j].numJumps > 0)
					{
						validMoves.add(moves[j]);
					}
					//else no jump is available and all moves are valid
					else if (maxJumps == 0)
					{
						validMoves.add(moves[j]);
					}
				}
			}
		}
		moves = new State[validMoves.size()];
		validMoves.toArray(moves);
		validMoves.clear();
		
		//if a jump is available for any checker piece,
		//eliminate any move that isn't a jump
		if (maxJumps > 0)
		{
			for (int i = 0; i < moves.length; i++)
			{
				if (moves[i].numJumps > 0)
				{
					moves[i].setTurn(!redMove);
					validMoves.add(moves[i]);
				}
			}
			moves = new State[validMoves.size()];
			validMoves.toArray(moves);
			validMoves.clear();
		}
		for (int i = 0; i < moves.length; i++)
		{
			moves[i].setTurn(!redMove);
		}
		return moves;
	}
	
	/**
	 * 
	 * @return true if it is red turn.
	 * @throws TurnNotSetException if the exception has not been set
	 */
	protected boolean isRedTurn()
	{
		if (turnSet == false)
		{
			throw new TurnNotSetException();
		}
		return itIsRedTurn;
	}

	/**
	 * Sets the turn
	 * @param redTurn true if it reds turn, false if white's
	 */
	protected void setTurn(boolean redTurn)
	{
		itIsRedTurn = redTurn;
		turnSet = true;
	}
	
	/**
	 * @param index the board position
	 * @return true if the board position at the index is red
	 */
	private boolean isRed(int index)
	{
		char c = board.charAt(index);
		return c == RED || c == RED_KING;
	}

	/**
	 * @param index the board position
	 * @return true if the board position at the index is white
	 */
	private boolean isWhite(int index)
	{
		char c = board.charAt(index);
		return c == WHITE || c == WHITE_KING;
	}

	/**
	 * @param index the board position
	 * @return true if the board position at the index is red or white king
	 */
	private boolean isKing(int index)
	{
		char c = board.charAt(index);
		return c == WHITE_KING || c == RED_KING;
	}

	/**
	 * @param index the board position
	 * @param onlyJumps if true, only the jump moves are returned.
	 * @return all the states that are successors for a single piece.
	 */
	private State[] getValidMoves(int index, boolean onlyJumps)
	{
		Vector list = new Vector(4);
		boolean red = isRed(index);
		boolean white = isWhite(index);
		boolean king = isKing(index);
		Position pos = getPosition(index);
		Position nextPosition;
		Position nextNextPosition;
		State[] tempMoves = new State[0];
		if (red || king)
		{
			// Move right
			nextPosition = new Position(pos.getX() + 1, pos.getY() - 1);
			nextNextPosition = new Position(pos.getX() + 2, pos.getY() - 2);
			tempMoves = determineMove(pos, nextPosition, nextNextPosition, red,
					onlyJumps);
			for (int i = 0; i < tempMoves.length; i++)
			{
				list.add(tempMoves[i]);
			}
			// Move left
			nextPosition = new Position(pos.getX() - 1, pos.getY() - 1);
			nextNextPosition = new Position(pos.getX() - 2, pos.getY() - 2);
			tempMoves = determineMove(pos, nextPosition, nextNextPosition, red,
					onlyJumps);
			for (int i = 0; i < tempMoves.length; i++)
			{
				list.add(tempMoves[i]);
			}
		}
		if (white || king)
		{
			// Move right
			nextPosition = new Position(pos.getX() + 1, pos.getY() + 1);
			nextNextPosition = new Position(pos.getX() + 2, pos.getY() + 2);
			tempMoves = determineMove(pos, nextPosition, nextNextPosition, red,
					onlyJumps);
			for (int i = 0; i < tempMoves.length; i++)
			{
				list.add(tempMoves[i]);
			}
			// Move left
			nextPosition = new Position(pos.getX() - 1, pos.getY() + 1);
			nextNextPosition = new Position(pos.getX() - 2, pos.getY() + 2);
			tempMoves = determineMove(pos, nextPosition, nextNextPosition, red,
					onlyJumps);
			for (int i = 0; i < tempMoves.length; i++)
			{
				list.add(tempMoves[i]);
			}
		}
		tempMoves = new State[list.size()];
		list.toArray(tempMoves);
		return tempMoves;
	}

	/**
	 * 
	 * @param pos
	 * @param nextPosition
	 * @param nextNextPosition
	 * @param redPiece
	 * @param onlyJumps
	 * @return the states resulting from a move or jump.  Recursively calls 
	 * getValidMoves() on jumps to determine if multiple jumps are possible.  
	 */
	private State[] determineMove(Position pos, Position nextPosition,
			Position nextNextPosition, boolean redPiece, boolean onlyJumps)
	{
		int nextPositionLoc = getLocation(nextPosition);
		int nextNextPositionLoc;
		char nextNextPiece;
		char nextPosPiece;
		char thisPosPiece;
		State[] moves = new State[0];
		String tempBoard;
		State tempState;
		Integer[] theBoardMoves;
		ArrayList list;
		if (nextPositionLoc != -1)
		{
			list = new ArrayList();
			if (movesMade.length >= 2)
			{
				for (int i = 0; i < movesMade.length - 1; i++)
				{
					list.add(movesMade[i]);
				}
			}
			nextPosPiece = board.charAt(nextPositionLoc);
			int currentPositionLoc = getLocation(pos);
			nextNextPositionLoc = getLocation(nextNextPosition);
			thisPosPiece = board.charAt(currentPositionLoc);
			list.add(new Integer(currentPositionLoc));
			if (nextPosPiece == BLANK)
			{
				if (onlyJumps == false)
				{
					tempBoard = movePiece(board, currentPositionLoc,
							nextPositionLoc, thisPosPiece);
					moves = new State[1];
					list.add(new Integer(nextPositionLoc));
					theBoardMoves = new Integer[list.size()];
					list.toArray(theBoardMoves);
					moves[0] = new State(tempBoard, 0, theBoardMoves);
				}
			}
			else if ((redPiece && (nextPosPiece == WHITE || nextPosPiece == WHITE_KING))
					|| (!redPiece && (nextPosPiece == RED || nextPosPiece == RED_KING)))
			{
				if (nextNextPositionLoc != -1)
				{
					nextNextPiece = board.charAt(nextNextPositionLoc);
					if (nextNextPiece == BLANK)
					{
						//jumpPiece
						tempBoard = jumpPiece(board, currentPositionLoc,
								nextPositionLoc, nextNextPositionLoc,
								thisPosPiece);
						list.add(new Integer(nextNextPositionLoc));
						theBoardMoves = new Integer[list.size()];
						list.toArray(theBoardMoves);
						if (tempBoard.length() > BOARD_SIZE)
						{
							tempBoard = new String(tempBoard.substring(0,
									BOARD_SIZE));
							moves = new State[1];
							moves[0] = new State(tempBoard, numJumps + 1,
									theBoardMoves);
						}
						else
						{
							tempState = new State(tempBoard, numJumps + 1,
									theBoardMoves);

							moves = tempState.getValidMoves(
									nextNextPositionLoc, true);
							if (moves.length == 0)
							{
								moves = new State[1];
								moves[0] = tempState;
							}
						}
					}
				}
			}
		}
		return moves;
	}
	
	/**
	 * @param theBoard the current board
	 * @param currentPositionLoc the piece to move location on board
	 * @param nextPositionLoc the position on board to move to
	 * @param thisPositionPiece the actual piece 
	 * @return the new checker board after a move
	 */
	private String movePiece(String theBoard, int currentPositionLoc,
			int nextPositionLoc, char thisPositionPiece)
	{
		char[] newBoard;
		newBoard = theBoard.toCharArray();
		if (thisPositionPiece == RED && nextPositionLoc >= 29
				&& nextPositionLoc <= 32)
		{
			thisPositionPiece = RED_KING;
		}
		if (thisPositionPiece == WHITE && nextPositionLoc >= 1
				&& nextPositionLoc <= 4)
		{
			thisPositionPiece = WHITE_KING;
		}
		newBoard[currentPositionLoc] = BLANK;
		newBoard[nextPositionLoc] = thisPositionPiece;
		return new String(newBoard);
	}
	
	/**
	 * 
	 * @param theBoard the current board
	 * @param currentPositionLoc the piece to move location on board
	 * @param nextPositionLoc the piece location to be jumped on board
	 * @param thisPositionPiece the actual piece doing the jump
	 * @param nextNextPositionLoc
	 * @return the new checker board after a single jump
	 */
	private String jumpPiece(String theBoard, int currentPositionLoc,
			int nextPositionLoc, int nextNextPositionLoc, char thisPositionPiece)
	{
		char[] newBoard;
		newBoard = theBoard.toCharArray();
		boolean becameAKing = false;
		String returnBoard;
		if (thisPositionPiece == RED && nextNextPositionLoc >= 29
				&& nextNextPositionLoc <= 32)
		{
			thisPositionPiece = RED_KING;
			becameAKing = true;
		}
		if (thisPositionPiece == WHITE && nextNextPositionLoc >= 1
				&& nextNextPositionLoc <= 4)
		{
			thisPositionPiece = WHITE_KING;
			becameAKing = true;
		}
		newBoard[currentPositionLoc] = BLANK;
		newBoard[nextPositionLoc] = BLANK;
		newBoard[nextNextPositionLoc] = thisPositionPiece;
		returnBoard = new String(newBoard);
		//System.out.println(returnBoard+"\t"+returnBoard.length());
		if (becameAKing)
		{
			//add an extra character so that we can know that it became a king
			//and cannot continue to jump in a multiple jump situation.
			returnBoard += "$";
		}
		// System.out.println(returnBoard+"\t"+returnBoard.length());
		return returnBoard;
	}

	/**
	 * Determines a x-y coordinate for the board 1-32 positions.
	 */
	public Position getPosition(int loc)
	{
		int x;
		int y;
		switch (loc)
		{
			case 1 :
				x = 1;
				y = 7;
				break;
			case 2 :
				x = 3;
				y = 7;
				break;
			case 3 :
				x = 5;
				y = 7;
				break;
			case 4 :
				x = 7;
				y = 7;
				break;
			case 5 :
				x = 0;
				y = 6;
				break;
			case 6 :
				x = 2;
				y = 6;
				break;
			case 7 :
				x = 4;
				y = 6;
				break;
			case 8 :
				x = 6;
				y = 6;
				break;
			case 9 :
				x = 1;
				y = 5;
				break;
			case 10 :
				x = 3;
				y = 5;
				break;
			case 11 :
				x = 5;
				y = 5;
				break;
			case 12 :
				x = 7;
				y = 5;
				break;
			case 13 :
				x = 0;
				y = 4;
				break;
			case 14 :
				x = 2;
				y = 4;
				break;
			case 15 :
				x = 4;
				y = 4;
				break;
			case 16 :
				x = 6;
				y = 4;
				break;
			case 17 :
				x = 1;
				y = 3;
				break;
			case 18 :
				x = 3;
				y = 3;
				break;
			case 19 :
				x = 5;
				y = 3;
				break;
			case 20 :
				x = 7;
				y = 3;
				break;
			case 21 :
				x = 0;
				y = 2;
				break;
			case 22 :
				x = 2;
				y = 2;
				break;
			case 23 :
				x = 4;
				y = 2;
				break;
			case 24 :
				x = 6;
				y = 2;
				break;
			case 25 :
				x = 1;
				y = 1;
				break;
			case 26 :
				x = 3;
				y = 1;
				break;
			case 27 :
				x = 5;
				y = 1;
				break;
			case 28 :
				x = 7;
				y = 1;
				break;
			case 29 :
				x = 0;
				y = 0;
				break;
			case 30 :
				x = 2;
				y = 0;
				break;
			case 31 :
				x = 4;
				y = 0;
				break;
			case 32 :
				x = 6;
				y = 0;
				break;
			default :
				throw new IllegalArgumentException("Invalid Board Position = "
						+ loc);
		}
		return new Position(x, y);
	}

	/**
	 * Determines the board 1-32 positions based on a x-y coordinate position.
	 */
	public int getLocation(Position pos)
	{
		int x = pos.getX();
		int y = pos.getY();
		int answer = -1;
		if (x == 0)
		{
			if (y == 0)
				answer = 29;
			else if (y == 2)
				answer = 21;
			else if (y == 4)
				answer = 13;
			else if (y == 6)
				answer = 5;
		}
		else if (x == 1)
		{
			if (y == 1)
				answer = 25;
			else if (y == 3)
				answer = 17;
			else if (y == 5)
				answer = 9;
			else if (y == 7)
				answer = 1;
		}
		else if (x == 2)
		{
			if (y == 0)
				answer = 30;
			else if (y == 2)
				answer = 22;
			else if (y == 4)
				answer = 14;
			else if (y == 6)
				answer = 6;
		}
		else if (x == 3)
		{
			if (y == 1)
				answer = 26;
			else if (y == 3)
				answer = 18;
			else if (y == 5)
				answer = 10;
			else if (y == 7)
				answer = 2;
		}
		else if (x == 4)
		{
			if (y == 0)
				answer = 31;
			else if (y == 2)
				answer = 23;
			else if (y == 4)
				answer = 15;
			else if (y == 6)
				answer = 7;
		}
		else if (x == 5)
		{
			if (y == 1)
				answer = 27;
			else if (y == 3)
				answer = 19;
			else if (y == 5)
				answer = 11;
			else if (y == 7)
				answer = 3;
		}
		else if (x == 6)
		{
			if (y == 0)
				answer = 32;
			else if (y == 2)
				answer = 24;
			else if (y == 4)
				answer = 16;
			else if (y == 6)
				answer = 8;
		}
		else if (x == 7)
		{
			if (y == 1)
				answer = 28;
			else if (y == 3)
				answer = 20;
			else if (y == 5)
				answer = 12;
			else if (y == 7)
				answer = 4;
		}
		return answer;
	}

	/**
	 * @return the utiltity value of the state based on our heuristic
	 * Heuristic is based on
	 * 1) difference between opponents pieces
	 */
	public int calcUtility()
	{
		int utility;
		int numWhite = 0;
		int numRed = 0;
		//int numRedOnBorders = 0;
		//int numWhiteOnBorders = 0;
		int sumDistances = 0;
		boolean red;
		boolean white;
		Position pos1;
		Position pos2;
		
		
		//loops through a calculates the total distance (max x or y absolute difference)
		//for all pieces.  The goal is for this to be smallest; keeping the pieces close together.
		for (int i = START_POSITION; i < BOARD_SIZE; i++)
		{
			red = isRed(i);
			white = isWhite(i);
			
			
			for(int j=START_POSITION; j< BOARD_SIZE; j++)
			{
				if( (red && isWhite(j)) || (white && isRed(j)) ||
						(red && isRed(j)) || (white && isWhite(j)) )
				{
					pos1 = getPosition(i);
					pos2 = getPosition(j);
					sumDistances+= Math.max(Math.abs(pos1.getX()-pos2.getX()),
							Math.abs(pos1.getY()-pos2.getY()));
				}
			}
			/*
						
			if (red)
			{
				numRed++;
				if (onBorder(i))
				{
					numRedOnBorders++;
				}
			}
			if (white)
			{
				numWhite++;
				if (onBorder(i))
				{
					numWhiteOnBorders++;
				}
			}
			*/
		}
		if (agentIsRed)
		{
			//the multiplier of 10000 puts priority on having less than the opponent
			//want smaller totaldistance so subtract value
			//want smaller total number of pieces so subtract value
			utility = (numWhite - numRed)* 1000000  - sumDistances - (myNumPieces()+otherNumPieces());
			
			//adds one to utility if is your turn
			if(itIsRedTurn)
			{
				utility++;
			}
			
		}
		else
		{
			utility = (numRed - numWhite) * 1000000 - sumDistances - (myNumPieces()+otherNumPieces()); 
			//*100 + (numRedOnBorders-numWhiteOnBorders);
			
			if(!itIsRedTurn)
			{
				utility--;
			}
			
		}

		return utility;
	}

	/**
	 * @return true if the piece at the index is on an edge border square
	 */
	private boolean onBorder(int loc)
	{
		boolean answer = false;
		if ((loc >= 1 && loc <= 5) || (loc >= 28 && loc <= 32)
				|| (loc >= 12 && loc <= 13) || (loc >= 20 && loc <= 21))
		{
			answer = true;
		}
		return answer;
	}

	
	/**
	 * Displays the checkerboard
	 */
	public void displayBoard()
	{
		Position pos;
		int loc;
		for (int i = 7; i >= 0; i--)
		{
			System.out.println("==================================");
			for (int j = 0; j < 8; j++)
			{
				pos = new Position(j, i);
				loc = getLocation(pos);
				if (loc == -1)
				{
					System.out.print("|~~~");
				}
				else
				{
					System.out.print("| " + board.charAt(loc) + " ");
				}
			}
			System.out.println(" |");
		}
		System.out.println("==================================");
		if (movesMade.length > 0)
		{
			if (itIsRedTurn)
			{
				System.out.print("Move Made by WHITE was ==>  ");
			}
			else
			{
				System.out.print("Move Made by RED was ==>  ");
			}
			int start;
			for (int i = 0; i < movesMade.length; i++)
			{
				System.out.print(movesMade[i]);
				if (i < movesMade.length - 1)
				{
					System.out.print("-");
				}
			}
			System.out.println();
		}
		if (itIsRedTurn)
		{
			System.out.println("Turn ==> RED\n");
		}
		else
		{
			System.out.println("Turn ==> WHITE\n");
		}
		
	}
	
	/**
	 * @return the string representing the board
	 */
	public String getBoard()
	{
		return board;
	}

	/**
	 * Not implemented.  Intended to check user input if it was valid
	 */
	public boolean isMoveValid(int[] move)
	{
		return true;
	}

	/**
	 * Creates a state/board reflecting the sequence of moves
	 * in the inptu array
	 */
	public State makeOpponentMove(int[] move)
	{
		State newState = null;
		char[] newBoard = board.toCharArray();
		int diff;
		char thisPositionPiece;
		Position[] locs = new Position[move.length];
		for (int i = 0; i < locs.length; i++)
		{
			locs[i] = getPosition(move[i]);
		}

		//change the board
		for (int i = 0; i < locs.length - 1; i++)
		{
			thisPositionPiece = newBoard[move[i]];

			diff = Math.abs(locs[i].getY() - locs[i + 1].getY());
			if (diff == 1)
			{
			}
			else if (diff == 2)
			{
				int diffY = locs[i + 1].getY() - locs[i].getY();
				int diffX = locs[i + 1].getX() - locs[i].getX();
				int jumpX;
				int jumpY;
				if (diffY > 0)
				{
					jumpY = locs[i].getY() + 1;
				}
				else
				{
					jumpY = locs[i].getY() - 1;
				}
				if (diffX > 0)
				{
					jumpX = locs[i].getX() + 1;
				}
				else
				{
					jumpX = locs[i].getX() - 1;
				}
				newBoard[getLocation(new Position(jumpX, jumpY))] = BLANK;
			}
			if (thisPositionPiece == RED && move[i + 1] >= 29
					&& move[i + 1] <= 32)
			{
				thisPositionPiece = RED_KING;
			}
			if (thisPositionPiece == WHITE && move[i + 1] >= 1
					&& move[i + 1] <= 4)
			{
				thisPositionPiece = WHITE_KING;
			}
			newBoard[move[i + 1]] = thisPositionPiece;
			newBoard[move[i]] = BLANK;
		}
		movesMade = new Integer[move.length];
		for (int i = 0; i < movesMade.length; i++)
		{
			movesMade[i] = new Integer(move[i]);
		}
		newState = new State(new String(newBoard));
		
		//change turn to the agent
		newState.setTurn(!isRedTurn());
		//return new State with the new board
		return newState;
	}

	/*
	 * Sets the true if the agent is to be red
	 */
	public static void setAgentRed(boolean redAgent)
	{
		agentIsRed = redAgent;
	}

	/**
	 * @return the number of jumps made that resulted in this state
	 */
	public int getNumJumps()
	{
		return numJumps;
	}

	/**
	 * @return the sequence ov moves made by the agent to create this board.
	 * Returns nothing if the user move resulting board.
	 */
	protected Integer[] getMadeMoves()
	{
		return movesMade;
	}
	
	/**
	 * @return the number of pieces on the board for the color who's turn it is.
	 */
	public int myNumPieces()
	{
		int count =0;
		
		for(int i=START_POSITION; i<BOARD_SIZE; i++)
		{
			if( (isRedTurn() && isRed(i)) || (!isRedTurn() && isWhite(i)) )
			{
				count++;
			}
		}
		return count;
	}
	
	/**
	 * @return the number of pieces on the board for the color who's turn it isn't.
	 */
	public int otherNumPieces()
	{
		int count =0;
		
		for(int i=START_POSITION; i<BOARD_SIZE; i++)
		{
			if( (!isRedTurn() && isRed(i)) || (isRedTurn() && isWhite(i)) )
			{
				count++;
			}
		}
		return count;
	}
}
