/**
 * Union College CSc-244 AI Project 3 - CheckerBoard 
 * CheckerBoard is a class defines the gameboard and piece moves for a game of checkers. It 
 * also incorporates the logic for the agent and the alpha-beta pruning algorithm. 
 * 
 * @author  RJ Conti and Tom Bush
 * @version 03.11.2004
 */

import java.util.*;

public class CheckerBoard
{
    /*************
     * CONSTANTS *
     *************/
    final int MAX_INT      = 65536;  // integer for Min heuristic calculation
    final int MIN_INT      = -65536; // integer for Max heuristic calculation
    final int STARTLEVEL   = 0;      // top level of the tree
    final double TIMELIMIT = 55000;  // 55 seconds in milliseconds, agent cannot iterate 
                                     //  MinMax longer than this time
    final int STOPLEVEL     = 8;     // Level the agent should stop at
    
    
    /**********************
     * INSTANCE VARIABLES *
     **********************/
    private char[] checkerBoard;     // The game's checker board
    private char[] undoBoard1;       // Interim board for the undo calculation
    private char[] undoBoard2;       // The previous CB that can be brought back
    private int  boardSize = 8;      // Square board side size
    private char rChecker  = 'r';    // Red checker symbol
    private char rKing     = 'R';    // Red king symbol
    private char wChecker  = 'w';    // White checker symbol
    private char wKing     = 'W';    // White king symbol
    private char canMove   = ' ';    // Position checker can move symbol
    private char cantMove  = '-';    // Position checker can't move symbol
    private int goesFirst;           // Who goes first, 0 = agent, 1 = human
    private int agentColor;          // Pointer for agent's color: red or white
    private int whoseTurn = 0;       // Pointer for who's turn it is
    private int moveCounter = 0;     // Counts the number of moves, a move is 2 plies
    private long searchTime;         // Time the agent's move starts
    private int redPieces = 12;      // The number of red pieces left
    private int[] redLocs;           // Locations for the remaining red pieces
    private int whitePieces = 12;    // The number of white pieces left
    private int[] whiteLocs;         // Locations of the remaining white pieces
    private int resolution;          // How the game ended:
                                     //  0 = game is still ongoing
                                     //  1 = agent has no pieces left, agent wins
                                     //  2 = player has no pieces left, player wins
                                     //  3 = agent has no valid moves, player wins
                                     //  4 = player has no valid moves, agent wins
                                     //  5 = game prematurely halted by player
    int[] pointsTable = new int[]{ 1 , 1 , 1 , 1,
                                 1 , 2 , 2 , 2,
                                   2 , 3 , 3 , 1,
                                 1 , 3 , 4 , 2,
                                   2 , 4 , 3 , 1,
                                 1 , 3 , 3 , 2,
                                   2 , 2 , 2 , 1,
                                 1 , 1 , 1 , 1 };

    /***************
     * CONSTRUCTOR *
     ***************/
    public CheckerBoard()
    {   // initialize board for start of new game
        checkerBoard = new char[]{ rChecker, rChecker, rChecker, rChecker,
                                   rChecker, rChecker, rChecker, rChecker,
                                   rChecker, rChecker, rChecker, rChecker,
                                   canMove, canMove, canMove, canMove,
                                   canMove, canMove, canMove, canMove,
                                   wChecker, wChecker, wChecker, wChecker,
                                   wChecker, wChecker, wChecker, wChecker,
                                   wChecker, wChecker, wChecker, wChecker };
                                   
        // initialize a board for the undo funciton                           
        undoBoard1  =  new char[]{ rChecker, rChecker, rChecker, rChecker,
                                   rChecker, rChecker, rChecker, rChecker,
                                   rChecker, rChecker, rChecker, rChecker,
                                   canMove, canMove, canMove, canMove,
                                   canMove, canMove, canMove, canMove,
                                   wChecker, wChecker, wChecker, wChecker,
                                   wChecker, wChecker, wChecker, wChecker,
                                   wChecker, wChecker, wChecker, wChecker };
                                   
        // Due to the way the undo function works, an intermediary board is
        //  needed. 
        undoBoard2  =  new char[]{ rChecker, rChecker, rChecker, rChecker,
                                   rChecker, rChecker, rChecker, rChecker,
                                   rChecker, rChecker, rChecker, rChecker,
                                   canMove, canMove, canMove, canMove,
                                   canMove, canMove, canMove, canMove,
                                   wChecker, wChecker, wChecker, wChecker,
                                   wChecker, wChecker, wChecker, wChecker,
                                   wChecker, wChecker, wChecker, wChecker };
    } // end CheckerBoard() Constructor

    /***********
     * METHODS *
     ***********/
    /**
     * playGame()   This method is the meat of the program. It controls which
     *              player goes when. It updates the board, calles the 
     *              alpha-beta routine for the agent's move calculation and
     *              determines when the game is over. 
     * 
     * @param       n/a
     * @return      n/a 
     */
    public void playGame()
    {
        // Determine the locations of the red and white pieces on the gameboard and 
        //  save their locations in a separate array.
        arrayLocs starter = findPieceArray(checkerBoard);
        redLocs   = starter.getRedArray();
        whiteLocs = starter.getWhiteArray();
        
        // Get some preliminary information
        ReadStdinInt input = new ReadStdinInt();

        // Let user input which player goes first
        System.out.println("Who goes first? (0 = agent, 1 = human)");
        goesFirst = input.readInt();
        
        // Let user input the color of the agent, where the agent is the computer. 
        System.out.println("What color is the agent? (0 = red, 1 = white)");
        agentColor = input.readInt();
        
        printBoard();               // Print the starting board
        
        while (resolution == 0)     // Game is still ongoing
        {

            if ((whoseTurn%2) == goesFirst)     // Agent's turn
            {
                /**************
                 * AGENT TURN *
                 **************/
                System.out.print("Agent's Turn: ");
                if (agentColor == 0)
                    System.out.println("RED");
                else
                    System.out.println("WHITE");
                
                // A generic node to hold successor deternmined by the alpha-beta algorithm
                aNode uNode = new aNode();
                
                // Determine the start time so the agent does not take more
                //  than 1 minute to complete its search.
                searchTime = new Date().getTime();
                // Actual move calculation, start the alpha-beta calculator
                uNode = AlphaBetaSearch();
                if (uNode == null)      // Agent has no valid moves left, it wins
                {   
                    resolution = 3;     // Agent wins
                } else {
                    // Update the checkerboard with the state returned from the successor
                    checkerBoard = uNode.getBoard();  
                    // Get the list of moves to output to the screen for the user
                    Vector moves = uNode.getMoves();
                    // System.out.println("Move Heuristic: "+uNode.getH());
                    
                    // Print information to the user on the location of the piece moved 
                    //  and where it is moving/jumping to.
                    int move0 = ((Integer)moves.get(0)).intValue();
                    System.out.print("Move "+undoBoard1[move0]+" at "+(move0+1));
                    System.out.println(" according to the following move sequence: ");
                    for (int i = 0; i < moves.size(); i++) 
                    {
                        int temp = ((Integer)moves.get(i)).intValue()+1;
                        if (i != ((moves.size())-1))
                            System.out.print(temp + " - ");
                        else
                            System.out.println(temp);
                    }   // end FOR
                    printBoard();
                    
                    // See if agent's moves have left player with no valid moves
                    // Create a stack and find the player's successors 
                    Stack s = new Stack();
                    succStates sS = new succStates();
                    s = sS.findSuccs(checkerBoard, ((agentColor+1)%2));
                    // If the player has no more moves available, player wins
                    if (s.size() == 0)
                        resolution = 4;     // Player wins
                    
                    // Update the piece locations and amounts
                    arrayLocs temp = findPieceArray(checkerBoard);
                    whiteLocs   = temp.getWhiteArray();
                    redLocs     = temp.getRedArray();
                    whitePieces = whiteLocs.length;
                    redPieces   = redLocs.length;
    
                    if (agentColor == 0)    // If agent color is red
                    {
                        if (whitePieces == 0)   // Player has no pieces left
                            resolution = 2;     // Player wins
                    } else {
                        if (redPieces == 0)     // Player has no pieces left
                            resolution = 2;     // Player wins
                    }   // end IF
                                        
                    updateUndoArray(true);  // Update the undo arrays.
                    whoseTurn++;    // Toggle who's turn it is.
                    
                    // Increment moveCounter here only if the human went first.
                    if (goesFirst == 1)
                        moveCounter++;
                }   // end IF
            
            } else {                // Human's Turn
                
             /**************
             * HUMAN TURN *
             **************/
             
                System.out.print("Your Turn: ");
                if (agentColor == 1)
                    System.out.println("RED");
                else
                    System.out.println("WHITE");      
                    
                // Create an array to hold the moves input by the user
                int[] moves = new int[12];
                for (int j = 0; j < 12; j++)    // Seed the array
                    moves[j] = -1;
                int i = 0;
                
                // Prompt the user for the start of the move sequence.
                System.out.print("Which piece would you like to move?  ");
                int choice = input.readInt();

                while (choice != 0)  
                {   // 0 is the sentinel to stop getting input moves
                    if ((choice < 0) || ((choice > 33) && (choice < 99)) || (choice > 99))
                    {   // Inputs in this range are not accepted by the agent.
                        System.out.print("Invalid entry: Please re-enter:   ");
                    } else if (choice == 33) { // 33 is the sentinal for an undo
                        updateUndoArray(false);     // Initiate an undo
                        System.out.println("Last move is being undone.");
                        printBoard();
                        // Since an undo can be done anytime, even after a move
                        //  has started, the move array has to be cleaned and
                        //  reinitialized.
                        moves = new int[12];
                        for (int j = 0; j < 12; j++)    // Seed the array
                            moves[j] = -1;
                        i = 0;
                    } else if (choice == 99) {   
                        // End the game prematurely.
                        resolution = 5;
                        break;
                    } else {
                        // Add the input to the move sequence.
                        moves[i] = choice;
                        i++;
                    }   // end IF

                // Get next move in the sequence
                System.out.print("Where would you like to move?  ");
                choice = input.readInt();
                }
                // Since the last number entered into the move array was 0,
                //  have to decrement i so that it the move routine does not
                //  use that as a valid board location.
                i--;
                
                int j = 0;      // Index for the movement of the user's piece
                while (j < i)   // Start updating the checkerboard
                {
                    int startMove = moves[j];   // Starting place
                    int endMove = moves[j+1];   // Ending place
                    // No movement (non-jump) between any board locations will
                    //  be greater than 5 away (legally anyway)
                    if (Math.abs(startMove-endMove) > 5) {
                        // It's a jump, update player's pieces
                        checkerBoard = jump(checkerBoard, startMove, endMove);
                        // Remove an agent's piece from the count
                        if (agentColor == 0)
                            redPieces--;
                        else
                            whitePieces--;
                    } else {
                        // It's just a move
                        checkerBoard = move(checkerBoard, startMove, endMove);
                    }   // end IF
                    j++;
                }   // end WHILE
                
                printBoard();     // Output the updated board to the screen
                
                // See if player's moves have left agent with no valid moves
                //  Create a stack and find the player's successors 
                Stack s = new Stack();
                succStates sS = new succStates();
                s = sS.findSuccs(checkerBoard, agentColor);
                // If the agent has no more moves available, agent wins
                if (s.size() == 0)
                    resolution = 3;     // Agent wins
                
                // Update the piece locations and amounts
                arrayLocs temp = findPieceArray(checkerBoard);
                whiteLocs   = temp.getWhiteArray();
                redLocs     = temp.getRedArray();
                whitePieces = whiteLocs.length;
                redPieces   = redLocs.length;
                
                if (agentColor == 0)        // If agent color is red
                {
                    if (redPieces == 0)     // Agent has no pieces left
                    {
                        resolution = 1;     // Agent wins
                    }   // end IF
                } else {
                    if (whitePieces == 0)   // Agent has no pieces left
                    {
                        resolution = 1;     // Agent wins
                    }   // end IF
                }   // end IF
                
                whoseTurn++;
                // Increment moveCounter here only if the agent went first
                if (goesFirst == 0)
                    moveCounter++;
            }   // end iF            
            
        }   // end WHILE
        
        System.out.println("resolution = " + resolution);
        gameStats();
    }   // end playGame()
    
    /**
     * AlphaBetaSearch  Starter function for the alpha-beta search aglorithm. This function
     *                  finds the successors for the current gameboard and determines the 
     *                  best successor using alpha-beta pruning. It will return the minimal
     *                  jump from the successors, if at least one jump is present, otherwise
     *                  it returns the best move of all possible movements.
     * 
     * @param       n/a
     * @return      aNode   a node which contains the successor board and the moves 
     */        
    private aNode AlphaBetaSearch()
    {
        int alpha = MIN_INT;    // Set alpha to the lowest value
        int beta  = MAX_INT;    // Set beta to the highest value
        int v = MIN_INT;        // Set v to the lowest value
        int jumps = 0;
        aNode nextAction = new aNode();     // Node to hold the move which will be the next
                                            //  action returned to the agent as the best move
        // Create a stack to hold the successor states 
        Stack sucks      = new Stack();
        // Create the object to find the states
        succStates succs = new succStates();
        sucks            = succs.findSuccs(checkerBoard, agentColor);
        
        aNode nextSuccessor = new aNode();  // A node to contain the next successor
        int size = sucks.size();    // Determine the number of successors in the stack
        if (size != 0)
        {
            for (int i = 0; i < size; i++)  // For each of the successor states
            {
                nextSuccessor = sucks.pop();
                // Determine if this successor has jumps. If so, compare it with the stored nextMove. 
                //  If they have the same number of moves, determine which has the highest heuristic.
                if ((nextSuccessor.getJumps() > 0) && (nextAction.getJumps() == 0))
                {
                    // Update the nextAction with the present move
                    nextAction.setBoard(nextSuccessor.getBoard());
                    nextAction.setMoves(nextSuccessor.getMoves());
                    nextAction.setJumps(nextSuccessor.getJumps());
                } else if (nextSuccessor.getJumps() == nextAction.getJumps()) {
                    // # of jumps the same, start the MiniMax algorithm on this branch and
                    //  determine which has the best heuristic
                    if ((MinValue(nextSuccessor.getBoard(), STARTLEVEL, alpha, beta)) > v)
                    {
                        // Update the nextAction with the present move
                        nextAction.setBoard(nextSuccessor.getBoard());
                        nextAction.setMoves(nextSuccessor.getMoves());
                        nextAction.setJumps(nextSuccessor.getJumps());
                    }   // end IF
    
                // if this is a shorter jump than the saved one
                } else if ((nextSuccessor.getJumps() < nextAction.getJumps()) && 
                           (nextSuccessor.getJumps() > 0)) {
                        // Update the nextAction with the present move
                        nextAction.setBoard(nextSuccessor.getBoard());
                        nextAction.setMoves(nextSuccessor.getMoves());
                        nextAction.setJumps(nextSuccessor.getJumps());
                }  // end IF
            }
        } else {
            nextAction = null;
        }   // end IF
        return nextAction;
    }   // end maxValue    
    
    /**
     * MinValue     maxValue function implemented straight out of the class text
     * 
     * @param       board   an array of characters representing the gameboard
     * @param       lev     the present level of the miniMax search
     * @param       alpha   the value of alpha for the a-b pruning
     * @param       beta    the value of beta for the a-b pruning
     * @return      int     the value of the smallest successor of this node 
     */    
    private int MaxValue(char[] board, int lev, int alpha, int beta)
    {
        int    color       = agentColor;             // Get the agent's color
        int    level       = lev+1;                  // Increment level
        double currentTime = new Date().getTime();   // Get present time
        int    v           = MIN_INT;                // The return value
    
        // Create a stack to hold the successor states 
        Stack sucks      = new Stack();
        // Create the object to find the states
        succStates succs = new succStates();
        sucks            = succs.findSuccs(board, color);
             
        int size = sucks.size();    // Determine the number of successors in the stack
        if ((level == STOPLEVEL) || ((currentTime - searchTime) > TIMELIMIT) || (size == 0)) 
        {       
            // Calculate the value of the present board configuration
            v = calcHeuristic(board);
        } else {    
            // For each of the successor states,
            for (int i = 0; i < size; i++)
            {   // Initiate a min function and take the maximum of v and the result 
                v = Math.max(v, MinValue(sucks.pop().getBoard(), level, alpha, beta));
                if (v >= beta)
                    return v;
                alpha = Math.max(alpha, v);
            }   // end FOR
        }   // end IF
        return v;
    }   // end maxValue
    
    /**
     * MinValue     minValue function implemented straight out of the class text
     * 
     * @param       board   an array of characters representing the gameboard
     * @param       lev     the present level of the miniMax search
     * @param       alpha   the value of alpha for the a-b pruning
     * @param       beta    the value of beta for the a-b pruning
     * @return      int     the value of the smallest successor of this node 
     */    
    private int MinValue(char[] board, int lev, int alpha, int beta)
    {
        int    color       = (agentColor+1)%2;       // Get the player's color
        int    level       = lev+1;                  // Increment level
        double currentTime = new Date().getTime();   // Get present time
        int    v           = MAX_INT;                // the return value
    
        // Create a stack to hold the successor states 
        Stack sucks      = new Stack();
        // Create the object to find the states
        succStates succs = new succStates();
        sucks            = succs.findSuccs(board, color);
             
        int size = sucks.size();        
        if ((level == STOPLEVEL) || ((currentTime - searchTime) > TIMELIMIT) || (size == 0)) 
        {        
            // Calculate the value of the present board configuration
            v = calcHeuristic(board);
        } else {    
            // For each of the successor states,
            for (int i = 0; i < size; i++)
            {   // Initiate a max function and take the maximum of v and the result 
                v = Math.min(v, MaxValue(sucks.pop().getBoard(), level, alpha, beta));
                if (v <= alpha)
                    return v;
                beta = Math.min(beta, v);
            }
        }   // end IF
        return v;
    }   // end maxValue
    
    /**
     * calcHeuristic() - Calculates the heuristic for a board configuration. 
     *                   Kings are worth more than a checker and the center
     *                   of the board is worth more because there are more 
     *                   ways to be jumped. Also, the direction of travel is 
     *                   biased for checkers and kings.
     * 
     * @param       downNode    
     * @return      int         
     */
    private int calcHeuristic(char[] board)
    {
        // Get information on the positions of the red and white pieces
        arrayLocs locations = findPieceArray(board);
        int[] redLocs   = locations.getRedArray();
        int[] whiteLocs = locations.getWhiteArray();
        int redHeuristic   = 0;     // Set the red heuristic to zero.
        int whiteHeuristic = 0;     // Set the white heurstic to zero.

        // Add the cost of each red piece to the overall red heuristic
        for (int i = 0; i < redLocs.length; i++)
        {
            if (board[redLocs[i]] == rKing)
            {
                // Kings are worth less points because they are bad
                redHeuristic = redHeuristic + 2*pointsTable[i] * (9-((i%4)+1));
            } else {
                redHeuristic = redHeuristic + pointsTable[i] * ((i%4)+1);
            }   // end IF
        }   // end FOR

        // Add the cost of each white piece to the overall white heuristic
        for (int i = 0; i < whiteLocs.length; i ++)
        {
            if (board[whiteLocs[i]] == wKing)
            {
                // Kings are worth less points because they are bad
                whiteHeuristic = whiteHeuristic + (2 * pointsTable[i] * (i%4)+1);
            } else {
                whiteHeuristic = whiteHeuristic + (pointsTable[i] * (9 - ((i%4)+1)));
            }   // end IF
        }   // end FOR
        
        int heuristic;
        // Calculate the overall state heuristic as player's heuristic - agent's heuristic.
        if (agentColor == 0)
            heuristic = whiteHeuristic - redHeuristic; //
        else
            heuristic = redHeuristic - whiteHeuristic; //
            
        return heuristic;
    }   // end calcHeuristic()
    
    /**
     * kingMe()
     * Crowns the specified piece a king.
     * 
     * @param       p    - the piece in the array to crown king
     * @return      n/a 
     */
    private char[] kingMe(char[] board, int p)
    {
        if (board[p] == rChecker)        // Red crowning
        {
            board[p] = rKing;
        } // end if rChecker
        else if (board[p] == wChecker)   // White crowning
        {
            board[p] = wKing;
        } // end else if wChecker
        else                             // Error trap
        {
            System.out.println("CheckerBoard::kingMe(" + p + ") - Error!");
        } // end else none of the above
        
        return board;
    } // end kingMe()

    /**
     * isEmpty()                Determines if a board position is empty.
     * 
     * @param       p           the position to test
     * @return      boolean     true if p is emty, else false 
     */
    private boolean isEmpty(char[] board, int p)
    {
        if (board[p] == canMove)
        {
            return true;
        } //end if empty
        else
        {
            System.out.println("CheckerBoard::isEmpty(" + p + ") = " +
                               checkerBoard[p] + " - Error!");
            return false;
        } // end not empty
    } // end isEmpty()

    /**
     * move()                   Moves a checker.
     * 
     * @param       start       the starting position of the moving piece
     *              finish      the ending position of the moving piece
     * @return      n/a 
     */
    private char[] move(char[] board, int start, int finish)
    {
        int f = finish - 1;      // array index offset
        int s = start - 1;      // array index offset
        
        if (isEmpty(board, f))
        {
            board[f] = board[s];
            board[s] = canMove;
        } // end if empty location

        // check for crowning possibilities
        if ( ((f == 0) || (f == 1) || (f == 2) || (f == 3)) && 
               board[f] == wChecker) // white crowning
        {
            board = kingMe(board, f);
        } // if white checker to be kinged
        else if ( ((f == 28) || (f == 29) || (f == 30) || (f == 31)) && 
               board[f] == rChecker) // red crowning
        {
            board = kingMe(board, f);
        } // else if red checker to be kinged      

        return board;
    } // end move()

    /**
     * jump()                   Jumps a checker (or king).
     * 
     * @param       start       the starting position of the moving piece
     *              finish      the ending position of the moving piece
     * @return      n/a 
     */    
    private char[] jump(char[] board, int start, int finish)
    {
        // Positions of the ending, starting and middle pieces in the jump
        int f = finish;
        int s = start;
        int jumped;
        
        // Call move to change the piece locations
        board = move(board, s, f); // Move the jumping piece
        
        // Now, remove the piece being jumped
        if ((1+((start-1)/4))%2 == 0) 
        {   // check for even row
            jumped = ((start+finish-1)/2);
        } else { // odd row
            jumped = ((start+finish+1)/2);
        }   // end IF
        // update agent's piece location and count
        board[jumped-1] = canMove;

        return board;
    }   // end Jump()
    
    /**
     * findPieceArray()         Determines where the pieces from each color are
     *                          on the board supplied. 
     * 
     * @param       char[]      The board configuration
     * @return      arrayLocs   The array of locations of all piece of 
     *                          each color
     */    
    private arrayLocs findPieceArray(char[] board)
    {
        // First go through and figure out how many of each color piece there are
        int redCount   = 0;
        int whiteCount = 0;
        for (int i = 0; i < 32; i++)
        {
            if ((board[i] == rChecker) || (board[i] == rKing)) 
            {
                redCount++;
            } else if ((board[i] == wChecker) || (board[i] == wKing)) {
                whiteCount++;
            }   // end IF
        }   // end FOR
        
        // Next go through again and update the color's location arrays
        int[] tempRedPieces   = new int[redCount];
        int[] tempWhitePieces = new int[whiteCount];
        int redIndex   = 0;
        int whiteIndex = 0;
        for (int i = 0; i < 32; i++)
        {
            if ((board[i] == rChecker) || (board[i] == rKing)) 
            {
                tempRedPieces[redIndex] = i;
                redIndex++;
            } else if ((board[i] == wChecker) || (board[i] == wKing)) {
                tempWhitePieces[whiteIndex] = i;
                whiteIndex++;
            }   // end IF
        }   // end FOR
        
        // Want to return two arrays, so have to pack the arrays in an arrayLocs datatype
        arrayLocs temp = new arrayLocs(tempWhitePieces, tempRedPieces);
        return temp;    
    }   // end findPieceArray()
    
    /**
     * updateUndoArray()        This procedure updates the two arrays used for 
     *                          the undo function. This procedure takes care of
     *                          two things, 
     *                          1) regenerates the previous board if an UNDO 
     *                             has been requested,
     *                          2) updates the undoBoards after each move 
     *                             (i.e., 1 ply by each player)
     * @param       boolean     a boolean value to determine if the board
     *                          should be updated or not
     * @return      n/a
     */  
    private void updateUndoArray(boolean update)
    {
        if (update == false)
        {
            // an undo has been requested
            for (int i = 0; i < 32; i++) 
            {
                // Restore the last checkerboard
                checkerBoard[i] = undoBoard2[i];
                // Update the arrays that contain the red adn white pieces
                arrayLocs temp = findPieceArray(checkerBoard);
                whiteLocs   = temp.getWhiteArray();
                redLocs     = temp.getRedArray();
                whitePieces = whiteLocs.length;
                redPieces   = redLocs.length;
            }   // end FOR
        } else {
            // update undo boards to current information
            for (int i = 0; i < 32; i++)
            {
                undoBoard2[i] = undoBoard1[i];
                undoBoard1[i] = checkerBoard[i];
            }   // end FOR
        }   // end IF
    }   // end updateArray()
    
    /**
     * printBoard()         Prints the checker board.
     * 
     * @param       n/a
     * @return      n/a 
     */
    private void printBoard()
    {
        int piece = 0;  // index into checkerBoard array for printing the 
                        // pieces
        
        System.out.println();
        System.out.println("               RED               ");
        System.out.println("---------------------------------");

        // print the checker board one row at a time
        for (int row= 0; row < boardSize; row++)
        {
            System.out.print("|");
            if ( (row + 1) % 2 == 1)  // the row is odd
            {
                for (int c = 0; c < boardSize; c++)
                {
                    if ( (c + 1) % 2 == 1)  // the column is odd
                    {
                        System.out.print(" " + cantMove + " |");
                    } // end if odd column
                    else                    // the column is even
                    {
                        System.out.print(" " + checkerBoard[piece] + " |");
                        piece++;
                    } // end else column even
                } // end for c
            } // end if odd row

            else                    // the row is even
            {
                for (int k = 0; k < boardSize; k++)
                {
                    if ( (k + 1) % 2 == 1)  // the column is odd
                    {
                        System.out.print(" " + checkerBoard[piece] + " |");
                        piece++;
                    } // end if
                    else                    // the column is even
                    {
                        System.out.print(" " + cantMove + " |");
                    } // end else column even
                } // end for k
            } // end else row even
            System.out.println();
            System.out.println("---------------------------------");
        } // end for row
        System.out.println("              WHITE              ");   
        System.out.println();
    } // end printBoard()        
        
    /**
     * gameStates   Outputs the stats for the game. Who won and the number of moves incurred.
     * 
     * @param       n/a
     * @return      n/a 
     */
    private void gameStats()
    {        
        System.out.println("MovementCounter = " + moveCounter);
        System.out.println();
        
        System.out.println(" GGGG   AAAA  M     M EEEEEE    OOOO  V     V  EEEEEE RRRRR ");
        System.out.println("G    G A    A MM   MM E        O    O V     V  E      R    R");
        System.out.println("G      A    A M M M M EEE      O    O V     V  EEE    R    R");
        System.out.println("G  GGG AAAAAA M  M  M E        O    O  V   V   E      RRRRR ");
        System.out.println("G    G A    A M     M E        O    O   V V    E      R    R");
        System.out.println(" GGG G A    A M     M EEEEEE    OOOO     V     EEEEEE R    R");
        System.out.println();
        
        switch(resolution) {
            case 1: System.out.println("Agent wins!");
                    break;
            case 2: System.out.println("Player wins!");
                    break;
            case 3: System.out.println("Agent wins! No more moves left");
                    break;
            case 4: System.out.println("Player wins! No more moves left");
                    break;
            case 5: System.out.println("Halted by player!");
                    break;
            default:
                    System.out.println("I dunno what happened!");
         }   // end switch
    } // end printBoard()
    
} // end CheckerBoard Class